/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

/* Test program for Steve.  Make a global read-only table available to
   each process by sending it from process 0 to one process per node,
   which shares it via MPI-3 shared memory with the other processes on
   that node.  There will be two shared arrays, one of ints, the other
   of doubles.  We call process 0 on each node the "leader."
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "mpi.h"

int main(int argc, char *argv[])
{
    int  topsize, toprank, nodesize, noderank, leadersize, leaderrank;
    int  i, color, key, intdispunit, doubledispunit;
    void *intbaseptr, *doublebaseptr;
    void *rank0_intbaseptr, *rank0_doublebaseptr;
    int  *intshmem;
    double *doubleshmem;
    MPI_Aint intmemsize, doublememsize;
    MPI_Comm topcomm, nodecomm, leadercomm;
    MPI_Win  intshwin, doubleshwin;
#define INTARRAYSIZE    100
#define DOUBLEARRAYSIZE 100
    int    intarray[INTARRAYSIZE];       /* stand-in for file of data to be shared */
    double doublearray[DOUBLEARRAYSIZE]; /* stand-in for file of data to be shared */

    MPI_Init(0, 0);
    
    MPI_Comm_dup(MPI_COMM_WORLD, &topcomm); /* to ensure generality */

    MPI_Comm_rank(topcomm, &toprank);
    MPI_Comm_size(topcomm, &topsize);

    /*  Initialize the data arrays on process 0 to represent the file data*/
    if (toprank == 0) {
        for (i = 0; i < INTARRAYSIZE; i++) intarray[i] = i;
        for (i = 0; i < DOUBLEARRAYSIZE; i++) doublearray[i] = (double) i;
    }

    /* split into subcommunicators, one per shared-memory node */
    key = 0;			/* rank order not important */
    MPI_Comm_split_type(topcomm, MPI_COMM_TYPE_SHARED, key, MPI_INFO_NULL, &nodecomm);

    MPI_Comm_rank(nodecomm, &noderank);
    MPI_Comm_size(nodecomm, &nodesize);
    printf("rank %d of %d in nodecomm is rank %d in topcomm\n", noderank, nodesize, toprank);

    /* Create a communicator composed of all the proc 0's on the nodes */
    if (noderank == 0)
        color = 0;
    else
        color = 1;              /* leadercomm is unused by these ranks */
    MPI_Comm_split(topcomm, color, key, &leadercomm);
    MPI_Comm_size(leadercomm, &leadersize);
    MPI_Comm_rank(leadercomm, &leaderrank);
    if (color == 0)
        printf("rank %d in leadercomm is rank %d in topcomm\n", leaderrank, toprank);

    /* Process 0 on each node allocates the whole shared memory windows */
    if (noderank == 0) {
        intmemsize    = INTARRAYSIZE * sizeof(int);
        doublememsize = DOUBLEARRAYSIZE * sizeof(double);
    }
    else {
        intmemsize = 0;
        doublememsize = 0;
    }

    /* Allocate two windows, one for each array */
    MPI_Win_allocate_shared(intmemsize,    MPI_INT, MPI_INFO_NULL, nodecomm, &intbaseptr,    &intshwin);
    MPI_Win_allocate_shared(doublememsize, MPI_DOUBLE, MPI_INFO_NULL, nodecomm, &doublebaseptr, &doubleshwin);

    printf("rank %d in nodecomm got intbaseptr %p from allocate_shared\n",    noderank, intbaseptr);
    printf("rank %d in nodecomm got doublebaseptr %p from allocate_shared\n", noderank, doublebaseptr);

    /* Get address of beginning of shared memory areas acquired by proc 0 */

    MPI_Win_shared_query(intshwin,    0, &intmemsize,    &intdispunit,    &rank0_intbaseptr);
    intshmem = (int *) rank0_intbaseptr; /* rename for convenience */
    MPI_Win_shared_query(doubleshwin, 0, &doublememsize, &doubledispunit, &rank0_doublebaseptr);
    doubleshmem = (double *) rank0_doublebaseptr; /* likewise */
    printf("rank %d has nodecomm 0's intbaseptr %p from shared query\n",     noderank, rank0_intbaseptr);
    printf("rank %d has nodecomm 0's doublebaseptr %p from shared quersy\n", noderank, rank0_doublebaseptr);

    /* Proc 0 in topcomm (also is 0 in leadercomm) copies the arrays into his own shared memory
       and then broadcasts the arrays to the other processes in leadercomm */

    if (toprank == 0) {         /* is also rank 0 in leadercomm */
        memcpy(rank0_intbaseptr,    &intarray,    sizeof(intarray));
        memcpy(rank0_doublebaseptr, &doublearray, sizeof(doublearray));
    }

    if (color == 0) {           /* I am a proc 0 in my nodecomm and hence I am in leadercomm */
        MPI_Bcast(rank0_intbaseptr,    intmemsize,    MPI_BYTE, 0, leadercomm);
        MPI_Bcast(rank0_doublebaseptr, doublememsize, MPI_BYTE, 0, leadercomm);
    }
    MPI_Barrier(topcomm);

    /* Test whether the memory is really shared */

    printf("rank %d sees intshmem[1] as %d\n", toprank, intshmem[1]);
    printf("rank %d sees doubleshmem[1] as %f\n", toprank, doubleshmem[1]);

    MPI_Finalize();

    return 0;
}  
