
/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 */

/* This program is designed to test thread safety of malloc/free */
/* step 1:  each thread allocates some memory and then frees memory allocated by another thread */



#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

#ifdef __APPLE__

#ifndef PTHREAD_BARRIER_H_
#define PTHREAD_BARRIER_H_

#include <pthread.h>
#include <errno.h>

typedef int pthread_barrierattr_t;
typedef struct
{
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    int count;
    int tripCount;
} pthread_barrier_t;


int pthread_barrier_init(pthread_barrier_t *barrier, const pthread_barrierattr_t *attr, unsigned int count)
{
    if(count == 0)
    {
        errno = EINVAL;
        return -1;
    }
    if(pthread_mutex_init(&barrier->mutex, 0) < 0)
    {
        return -1;
    }
    if(pthread_cond_init(&barrier->cond, 0) < 0)
    {
        pthread_mutex_destroy(&barrier->mutex);
        return -1;
    }
    barrier->tripCount = count;
    barrier->count = 0;

    return 0;
}

int pthread_barrier_destroy(pthread_barrier_t *barrier)
{
    pthread_cond_destroy(&barrier->cond);
    pthread_mutex_destroy(&barrier->mutex);
    return 0;
}

int pthread_barrier_wait(pthread_barrier_t *barrier)
{
    pthread_mutex_lock(&barrier->mutex);
    ++(barrier->count);
    if(barrier->count >= barrier->tripCount)
    {
        barrier->count = 0;
        pthread_cond_broadcast(&barrier->cond);
        pthread_mutex_unlock(&barrier->mutex);
        return 1;
    }
    else
    {
        pthread_cond_wait(&barrier->cond, &(barrier->mutex));
        pthread_mutex_unlock(&barrier->mutex);
        return 0;
    }
}

#endif // PTHREAD_BARRIER_H_
#endif // __APPLE__



typedef struct {
    pthread_t tid;
    int tnum;
    int arg;
    void * blob;              /* memory allocated and freed */
} thread_ent;

#define MAX_THREADS 70
#define BLOBSIZE    23

static thread_ent thread_tab[MAX_THREADS];
static int num_threads_created;
static int num_hellos;

pthread_barrier_t tbarr1;        /* barrier for created threads */
pthread_barrier_t tbarr2;        /* barrier for created threads */

void * thread_code(void * dummy_arg);

int main(int argc, char *argv[])
{

    int i, rc;

    if (argc != 3) {
        printf("usage: %s <num threads to create> <num hellos from each thread>\n", argv[0]);
        exit(99);
    } else {
        num_threads_created = atoi(argv[1]);
	num_hellos = atoi(argv[2]);
    }
    if (num_threads_created > MAX_THREADS) {
        printf("Can't specify more than %d threads to create\n", MAX_THREADS);
        exit(98);
    }
          
    pthread_barrier_init(&tbarr1, NULL, num_threads_created);
    pthread_barrier_init(&tbarr2, NULL, num_threads_created);

    for (i = 0; i < num_threads_created; i++) {
        rc = pthread_create(&thread_tab[i].tid, NULL, thread_code, NULL);
        /* printf("created thread %0x\n", (unsigned int) thread_tab[i].tid); */
        if (rc)
            printf("non_zero return code %d from pthread_create\n", rc);
        else {
            thread_tab[i].tnum = i;
            thread_tab[i].arg  = i + 1; /* currently unused */
        }
    }
    for (i = num_threads_created; i < MAX_THREADS; i++)
        thread_tab[i].tnum = -1;

    /* main thread waits here */

    for (i = 0; i < num_threads_created; i++)
        pthread_join(thread_tab[i].tid, NULL);

    return 0;
}

void * thread_code(void * dummy_arg)
{
    pthread_t my_thread_id;
    int i, j, my_thread_num;

    pthread_barrier_wait(&tbarr1);

    my_thread_id = pthread_self();
    /* printf("my thread id is %0x\n", (unsigned int) my_thread_id); */

    /* find my thread number (index in thread_tab) */
    for (i = 0; i < num_threads_created; i++)
        if (pthread_equal(thread_tab[i].tid, my_thread_id))
            my_thread_num = i;

    for (j = 0; j < num_hellos; j++)
        printf("Hello %d from thread %d\n", j, my_thread_num);

    thread_tab[my_thread_num].blob = malloc((size_t) BLOBSIZE);

    pthread_barrier_wait(&tbarr2);

    if (my_thread_num != 0)
        free(thread_tab[my_thread_num-1].blob);
    else
        free(thread_tab[num_threads_created-1].blob);
             
    return 0;
}
